@@ -1,5 +1,5 @@
 /*
- * $RCSId: xc/lib/fontconfig/src/fcinit.c,v 1.7 2002/08/22 07:36:44 keithp Exp $
+ * fontconfig/src/fcinit.c
  *
  * Copyright © 2001 Keith Packard
  *
@@ -7,15 +7,15 @@
  * documentation for any purpose is hereby granted without fee, provided that
  * the above copyright notice appear in all copies and that both that
  * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of Keith Packard not be used in
+ * documentation, and that the name of the author(s) not be used in
  * advertising or publicity pertaining to distribution of the software without
- * specific, written prior permission.  Keith Packard makes no
+ * specific, written prior permission.  The authors make no
  * representations about the suitability of this software for any purpose.  It
  * is provided "as is" without express or implied warranty.
  *
- * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
  * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
@@ -58,35 +58,51 @@ FcConfig *
 FcInitLoadConfig (void)
 {
     FcConfig	*config;
-    
+
     FcInitDebug ();
     config = FcConfigCreate ();
     if (!config)
-	return FcFalse;
-    
+	return NULL;
+
     if (!FcConfigParseAndLoad (config, 0, FcTrue))
     {
 	FcConfigDestroy (config);
 	return FcInitFallbackConfig ();
     }
-    
+
     if (config->cacheDirs && config->cacheDirs->num == 0)
     {
+	FcChar8 *prefix;
+	size_t plen;
+
 	fprintf (stderr,
 		 "Fontconfig warning: no <cachedir> elements found. Check configuration.\n");
 	fprintf (stderr,
 		 "Fontconfig warning: adding <cachedir>%s</cachedir>\n",
 		 FC_CACHEDIR);
+	prefix = FcConfigXdgCacheHome ();
+	plen = prefix ? strlen ((const char *)prefix) : 0;
+	if (!prefix)
+	    goto bail;
+	prefix = realloc (prefix, plen + 12);
+	if (!prefix)
+	    goto bail;
+	memcpy (&prefix[plen], FC_DIR_SEPARATOR_S "fontconfig", 11);
+	prefix[plen + 11] = 0;
 	fprintf (stderr,
-		 "Fontconfig warning: adding <cachedir>~/.fontconfig</cachedir>\n");
+		 "Fontconfig warning: adding <cachedir prefix=\"xdg\">fontconfig</cachedir>\n");
+
 	if (!FcConfigAddCacheDir (config, (FcChar8 *) FC_CACHEDIR) ||
-	    !FcConfigAddCacheDir (config, (FcChar8 *) "~/.fontconfig"))
+	    !FcConfigAddCacheDir (config, (FcChar8 *) prefix))
 	{
+	  bail:
 	    fprintf (stderr,
 		     "Fontconfig error: out of memory");
+	    free (prefix);
 	    FcConfigDestroy (config);
 	    return FcInitFallbackConfig ();
 	}
+	free (prefix);
     }
 
     return config;
@@ -139,7 +155,7 @@ FcFini (void)
     if (_fcConfig)
 	FcConfigDestroy (_fcConfig);
 
-    FcPatternFini ();
+    FcObjectFini ();
     FcCacheFini ();
     if (FcDebug() & FC_DBG_MEMORY)
 	FcMemReport ();
@@ -221,7 +237,7 @@ static struct {
     { "vstack" },
     { "attr" },
     { "pstack" },
-    { "staticstr" },
+    { "sharedstr" },
 };
 
 static int  FcAllocCount, FcAllocMem;
